//****************************************************************************
// Copyright (C) 2001,2002,2003  PEAK System-Technik GmbH
//
// linux@peak-system.com
// www.peak-system.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// Maintainer(s): Klaus Hitschler (klaus.hitschler@gmx.de)
//****************************************************************************

//****************************************************************************
//
// pcan_usb.c - the outer usb parts for pcan-usb support
//
// $Log: pcan_usb.c,v $
// Revision 1.2  2003/03/02 10:46:31  klaus
// merged USB thread into main path
//
// Revision 1.1.2.36  2003/02/25 20:23:37  klaus
// moved pcan_usb_start() from pcan_open() to pcan_device_open()
//
// Revision 1.1.2.35  2003/02/24 19:12:43  klaus
// added debug message at 'while .. schedule()'
//
// Revision 1.1.2.34  2003/02/23 19:33:06  klaus
// tried to fix "not possible to stop" bug
//
// Revision 1.1.2.33  2003/02/16 19:55:52  klaus
// USB Integration, first non public release
//
// Revision 1.1.2.32  2003/02/16 16:36:16  klaus
// pcan_usb_kernel.c returned to main modules
//
// Revision 1.1.2.31  2003/02/09 13:18:09  klaus
// modifications to support linux 2.2.19 kernels
//
// Revision 1.1.2.30  2003/02/09 13:18:09  klaus
// modifications to support linux 2.2.19 kernels
//
// Revision 1.1.2.29  2003/02/09 10:29:20  klaus
// code cleanup, Release_20030208_x
//
// Revision 1.1.2.28  2003/02/08 19:33:48  klaus
// cosmetic changes
//
// Revision 1.1.2.27  2003/02/08 17:32:43  klaus
// modified to use pcan_usb_kernel as prorietary module
//
// Revision 1.1.2.26  2003/02/05 23:12:19  klaus
// adapted to RedHat 7.2
//
// Revision 1.1.2.25  2003/01/29 20:34:20  klaus
// release_20030129_a and release_20030129_u released
//
//****************************************************************************

//****************************************************************************
// INCLUDES
#include <src/pcan_common.h>     // must always be the 1st include
#include <linux/stddef.h>        // NULL
#include <linux/errno.h>
#include <linux/slab.h>          // kmalloc()
#ifndef LINUX_22
#include <linux/module.h>        // MODULE_DEVICE_TABLE()
#endif
#include <linux/usb.h>

#include <src/pcan_usb.h>
#include <src/pcan_usb_kernel.h>

//****************************************************************************
// DEFINES
#define PCAN_USB_MINOR_BASE 32           // starting point of minors for USB devices
#define PCAN_USB_VENDOR_ID  0x0c72
#define PCAN_USB_PRODUCT_ID 0x000c

#define URB_READ_BUFFER_SIZE      1024   // buffer for read URB data (IN)
#define URB_READ_BUFFER_SIZE_OLD    64   // used length for revision < 6

//****************************************************************************
// GLOBALS
#ifndef LINUX_22
static struct usb_device_id pcan_usb_ids[] =
{
	{ USB_DEVICE(PCAN_USB_VENDOR_ID, PCAN_USB_PRODUCT_ID) },
	{ }						// Terminating entry
};

MODULE_DEVICE_TABLE (usb, pcan_usb_ids);
#endif

//****************************************************************************
// LOCALS
static u16 usb_devices = 0;        // the number of accepted usb_devices

//****************************************************************************
// CODE

//****************************************************************************
// get cyclic data from endpoint 2
static void pcan_usb_write_notify(purb_t purb);

static int pcan_usb_write(struct pcandev *dev)
{
  int err = 0;
  USB_PORT *u = &dev->port.usb;
  int nDataLength = u->Endpoint[3].wDataSz;

  DPRINTK(KERN_DEBUG "%s: pcan_usb_write()\n", DEVICE_NAME);

  // don't do anything with non-existent hardware
  if (!dev->ucPhysicallyInstalled)
    return -ENODEV;

  err = pcan_hw_Y(dev, u->pucWriteBuffer, &nDataLength);

  // fill the URB and submit
  if (nDataLength)
  {
    FILL_BULK_URB(u->write_data, u->usb_dev,
                  usb_sndbulkpipe(u->usb_dev, u->Endpoint[3].ucNumber),
                  u->pucWriteBuffer, nDataLength, pcan_usb_write_notify, dev);

    // start next urb
    if ((err = usb_submit_urb(u->write_data)))
    {
      dev->nLastError = err;
      dev->dwErrorCounter++; 
      DPRINTK(KERN_ERR "%s: pcan_usb_write() can't submit! (%d)",DEVICE_NAME, err);
    }
    else
      atomic_add(1, &dev->port.usb.active_urbs);
  }

  // it's no error if I can't get more data but still a packet was sent
  if ((err == -ENODATA) && (nDataLength))
    err = 0;

  return err;
}

//****************************************************************************
// notify functions for read and write data
static void pcan_usb_write_notify(purb_t purb)
{
  int err     = 0;
  u16 wwakeup = 0;
	struct pcandev *dev = purb->context;

  // DPRINTK(KERN_DEBUG "%s: pcan_usb_write_notify() (%d)\n", DEVICE_NAME, purb->status);

  // un-register outstanding urb
  atomic_sub(1, &dev->port.usb.active_urbs);

  // don't count interrupts - count packets
  dev->dwInterruptCounter++;                                

  // do write
  if (!purb->status) // stop with first error
    err = pcan_usb_write(dev);

  if (err == -ENODATA)
    wwakeup++;
  else
  {
    dev->nLastError = err;
    dev->dwErrorCounter++; 
    dev->wCANStatus |= CAN_ERR_QXMTFULL; // fatal error!
  }

  if (wwakeup)
  {
    atomic_set(&dev->DataSendReady, 1); // signal to write I'm ready
    wake_up_interruptible(&dev->write_queue);
  }
}

static void pcan_usb_read_notify(purb_t purb)
{
  int err = 0;
	struct pcandev *dev = purb->context;
  USB_PORT *u = &dev->port.usb;

  // DPRINTK(KERN_DEBUG "%s: pcan_usb_read_notify() (%d)\n", DEVICE_NAME, purb->status);

  // un-register outstanding urb
  atomic_sub(1, &dev->port.usb.active_urbs);

  // don't count interrupts - count packets
  dev->dwInterruptCounter++;                                

  // do interleaving read
  if (!purb->status && dev->ucPhysicallyInstalled) // stop with first error
  {
    u8  *m_pucTransferBuffer = purb->transfer_buffer;
    int m_nCurrentLength     = purb->actual_length;

    // buffer interleave to increase speed
    if (m_pucTransferBuffer == u->pucReadBuffer[0])
    {
      FILL_BULK_URB(purb, u->usb_dev,
                    usb_rcvbulkpipe(u->usb_dev, u->Endpoint[2].ucNumber),
                    u->pucReadBuffer[1], u->wReadBufferLength, pcan_usb_read_notify, dev);
    }
    else
    {
      FILL_BULK_URB(purb, u->usb_dev,
                    usb_rcvbulkpipe(u->usb_dev, u->Endpoint[2].ucNumber),
                    u->pucReadBuffer[0], u->wReadBufferLength, pcan_usb_read_notify, dev);
    }

    // start next urb
    if ((err = usb_submit_urb(purb)))
    {
      dev->nLastError = err; 
      dev->dwErrorCounter++;
      printk(KERN_ERR "%s: pcan_usb_read_notify() can't submit! (%d)",DEVICE_NAME, err);
    }
    else
      atomic_add(1, &dev->port.usb.active_urbs);

    err = pcan_hw_X(dev, m_pucTransferBuffer, m_nCurrentLength);
    if (err)
    {
      dev->nLastError = err; 
      dev->wCANStatus |=  CAN_ERR_QOVERRUN;
      dev->dwErrorCounter++;
      DPRINTK(KERN_DEBUG "%s: error %d from hw_DecodeMessage()\n", DEVICE_NAME, err);
    }
  }
  else
  {
    printk(KERN_ERR "%s: read data stream torn off caused by ", DEVICE_NAME);
    if (!dev->ucPhysicallyInstalled)
      printk("device plug out!\n");
    else
      printk("err %d!\n", purb->status);
  }
}

//****************************************************************************
// usb resource allocation 
//
static int pcan_usb_allocate_resources(struct pcandev *dev)
{
  int err = 0;
  USB_PORT *u = &dev->port.usb;

  DPRINTK(KERN_DEBUG "%s: pcan_usb_allocate_resources()\n", DEVICE_NAME);

  // allocate X data for comparison
  if (!u->pvXptr)
  {
    if (!(u->pvXptr = kmalloc(pcan_X(), GFP_KERNEL)))
    {
      err = -ENOMEM;
      goto fail;
    }
  }
  dev->wInitStep = 4;

  // make param URB
  u->param_urb = usb_alloc_urb(0);
  if (!u->param_urb)
    err = -ENOMEM;
  dev->wInitStep = 5;

  // allocate write buffer
  u->pucWriteBuffer =  kmalloc(u->Endpoint[3].wDataSz, GFP_KERNEL);
  DPRINTK(KERN_DEBUG "%s: kmalloc() = %p\n", DEVICE_NAME, u->pucWriteBuffer);
  if (!u->pucWriteBuffer)
  {
    err = -ENOMEM;
    goto fail;
  }
  dev->wInitStep = 6;

  // make write urb
  u->write_data = usb_alloc_urb(0);
  if (!u->write_data)
  {
    err = -ENOMEM;
    goto fail;
  }
  dev->wInitStep = 7;

  // reset telegram count
  u->dwTelegramCount = 0;

  // allocate both read buffers for URB
  u->pucReadBuffer[0] =  kmalloc(URB_READ_BUFFER_SIZE * 2, GFP_KERNEL);
  DPRINTK(KERN_DEBUG "%s: kmalloc() = %p\n", DEVICE_NAME, u->pucReadBuffer[0]);
  if (!u->pucReadBuffer[0])
  {
    err = -ENOMEM;
    goto fail;
  }
  u->pucReadBuffer[1] = u->pucReadBuffer[0] + URB_READ_BUFFER_SIZE;
  dev->wInitStep = 8;

  // make read urb
  u->read_data = usb_alloc_urb(0);
  if (!u->read_data)
  {
    err = -ENOMEM;
    goto fail;
  }
  dev->wInitStep = 9;

  // different revisions use different buffer sizes
  if (u->ucRevision < 6)
    u->wReadBufferLength = URB_READ_BUFFER_SIZE_OLD;
  else
    u->wReadBufferLength = URB_READ_BUFFER_SIZE;

  fail:
  return err;
}

//****************************************************************************
// usb resource de-allocation 
//
static int pcan_usb_free_resources(struct pcandev *dev)
{
  int err = 0;
  USB_PORT *u = &dev->port.usb;

  DPRINTK(KERN_DEBUG "%s: pcan_usb_free_resources()\n", DEVICE_NAME);

  // at this point no URB must be pending
  // this is forced at pcan_usb_stop
  switch (dev->wInitStep)
  {
    case 9:
      // free read URB
      usb_free_urb(u->read_data);

    case 8:
      kfree(u->pucReadBuffer[0]);

    case 7:
      // free write urb
      usb_free_urb(u->write_data);

    case 6:
      kfree(u->pucWriteBuffer);

    case 5:
      // free param urb
      usb_free_urb(u->param_urb);

    case 4:
      // remove X information
      if (u->pvXptr)
      {
        kfree(u->pvXptr);
        u->pvXptr = NULL;
      }

      dev->wInitStep = 3;
  }

  return err;
}

//****************************************************************************
// start and stop functions for CAN data IN and OUT
static int pcan_usb_start(struct pcandev *dev)
{
  int err = 0;
  USB_PORT *u = &dev->port.usb;

  DPRINTK(KERN_DEBUG "%s: pcan_usb_start()\n", DEVICE_NAME); 

  FILL_BULK_URB(u->read_data, u->usb_dev,
                usb_rcvbulkpipe(u->usb_dev, u->Endpoint[2].ucNumber),
                u->pucReadBuffer[0], u->wReadBufferLength, pcan_usb_read_notify, dev);

  // submit urb
  if ((err = usb_submit_urb(u->read_data)))
    printk(KERN_ERR "%s: pcan_usb_start() can't submit! (%d)\n",DEVICE_NAME, err);
  else
    atomic_add(1, &dev->port.usb.active_urbs);

  return err;
}

static int pcan_usb_stop(struct pcandev *dev)
{
  int err = 0;
  USB_PORT *u = &dev->port.usb;
  unsigned long dwFlags = 0;
  int i = 100;

  DPRINTK(KERN_DEBUG "%s: pcan_usb_stop()\n", DEVICE_NAME); 

  err = pcan_hw_SetCANOff(dev);

  // wait until all has settled
  mdelay(5);

  // unlink URB
  spin_lock_irqsave(&u->lock, dwFlags);
  if (u->read_data->status == -EINPROGRESS)
  {
    err = usb_unlink_urb(u->read_data);
    DPRINTK(KERN_DEBUG "%s: usb_unlink_urb(), read = %d\n", DEVICE_NAME, err);
  }
  spin_unlock_irqrestore (&u->lock, dwFlags);

  // unlink writeURB
  spin_lock_irqsave(&u->lock, dwFlags);
  if (u->write_data->status == -EINPROGRESS)
  {
    err = usb_unlink_urb(u->write_data);
    DPRINTK(KERN_DEBUG "%s: usb_unlink_urb(), write = %d\n", DEVICE_NAME, err);
  }
  spin_unlock_irqrestore (&u->lock, dwFlags);

  // unlink parameter urb
  spin_lock_irqsave(&u->lock, dwFlags);
  if (u->param_urb->status == -EINPROGRESS)
  {
    err = usb_unlink_urb(u->param_urb);
    DPRINTK(KERN_DEBUG "%s: usb_unlink_urb(), param = %d\n", DEVICE_NAME, err);
  }
  spin_unlock_irqrestore (&u->lock, dwFlags);

  // wait until all urbs returned to sender
  // (I hope) this should be no problem because all urb's are unlinked 
  while ((atomic_read(&u->active_urbs) > 0) && (i--))
    schedule();
  if (i <= 0)
  {
    DPRINTK(KERN_ERR "%s: have still active URBs: %d!\n", DEVICE_NAME, atomic_read(&u->active_urbs));
  }

  return err;
}

//****************************************************************************
// remove device resources 
//
static int pcan_usb_cleanup(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_cleanup()\n", DEVICE_NAME);

  if (dev)
  {
    pcan_usb_free_resources(dev);

    switch(dev->wInitStep)
    {
      case 3: usb_devices--;
              pcan_drv.wDeviceCount--;
      case 2: list_del(&dev->list);
      case 1: 
      case 0: dev->wInitStep = 0;
              kfree(dev);
    }
  }
  
  return 0;
}

// dummy entries for request and free irq
static int pcan_usb_req_irq(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_req_irq()\n", DEVICE_NAME);
  return 0;
}

static void pcan_usb_free_irq(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_free_irq()\n", DEVICE_NAME);

  // mis-used here for another purpose
  // pcan_usb_free_irq() calls when the last path to device just closing
  // and the device itself is already plugged out
  if ((dev) && (!dev->ucPhysicallyInstalled))
    pcan_usb_cleanup(dev);
}

// interface depended open and close
static int pcan_usb_open(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_open()\n", DEVICE_NAME);

  return 0;
}

static int pcan_usb_release(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_release()\n", DEVICE_NAME);

  return 0;
}

// emulated device access functions
// call is only possible if device exists
static int pcan_usb_device_open(struct pcandev *dev, u16 btr0btr1, u8 bExtended, u8 bListenOnly)
{
  int err = 0;

  DPRINTK(KERN_DEBUG "%s: pcan_usb_device_open()\n", DEVICE_NAME);

  // first action: turn CAN off
  if ((err = pcan_hw_SetCANOff(dev)))
    goto fail;

  if ((err = pcan_usb_start(dev)))
    goto fail;

  // init hardware specific parts
  if ((err = pcan_hw_Init(dev, btr0btr1, bListenOnly)))
    goto fail;

  // store extended mode (standard still accepted)
  dev->bExtended = bExtended;

  // take a fresh status
  dev->wCANStatus = 0;

  // copy from NT driver
  mdelay(20);

  // last action: turn CAN on
  if ((err = pcan_hw_SetCANOn(dev)))
    goto fail;

  fail:
  return err;
}

static void pcan_usb_device_release(struct pcandev *dev)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_device_release()\n", DEVICE_NAME);

  // do not stop usb immediately, give a chance for the USB-dongle to send this telegram
  mdelay(100);

  pcan_usb_stop(dev);
}

static int pcan_usb_device_write(struct pcandev *dev)
{
  return pcan_usb_write(dev);
}

//****************************************************************************
// special assignment of minors due to PuP
//
static int assignMinorNumber(struct pcandev *dev)
{
  int searchedMinor;
  u8 occupied;
  struct pcandev   *devWork = (struct pcandev *)NULL;
  struct list_head *ptr;

  DPRINTK(KERN_DEBUG "%s: assignMinorNumber()\n", DEVICE_NAME);

  for (searchedMinor = PCAN_USB_MINOR_BASE; searchedMinor < (PCAN_USB_MINOR_BASE + 8); searchedMinor++)
  {
    occupied = 0;

    // loop trough my devices
    for (ptr = pcan_drv.devices.next; ptr != &pcan_drv.devices; ptr = ptr->next)
    {
      devWork = (struct pcandev *)ptr;  

      // stop if it is occupied
      if (devWork->nMinor == searchedMinor)
      {
        occupied = 1;
        break;
      }
    }
    
    // jump out when the first available number is found
    if (!occupied)
      break;
  }

  if (!occupied)
  {
    dev->nMinor = searchedMinor;
    return 0;
  }
  else
  {
    dev->nMinor = -1;
    return -ENXIO;
  }
}

//****************************************************************************
// things todo after plugin or plugout of device (and power on too)
//
#ifndef LINUX_22
static void *pcan_usb_plugin(struct usb_device *usb_dev, unsigned int interface, const struct usb_device_id *id_table)
#else
static void *pcan_usb_plugin(struct usb_device *usb_dev, unsigned int interface)
#endif
{
  struct pcandev *dev = NULL;
  int    err = 0;
  int    i;
  USB_PORT *u;
  struct usb_interface_descriptor *current_interface_setting;

  DPRINTK(KERN_DEBUG "%s: pcan_usb_plugin(0x%04x, 0x%04x, %d)\n", DEVICE_NAME, 
          usb_dev->descriptor.idVendor, usb_dev->descriptor.idProduct, interface);

  if ((usb_dev->descriptor.idVendor  == PCAN_USB_VENDOR_ID)  && 
      (usb_dev->descriptor.idProduct == PCAN_USB_PRODUCT_ID))
  {
    // take the 1st configuration (it's default)
    if (usb_set_configuration (usb_dev, usb_dev->config[0].bConfigurationValue) < 0) 
    {
      printk(KERN_ERR "%s: usb_set_configuration() failed!\n", DEVICE_NAME);
      goto reject;
    }

    // only 1 interface is supported
    if (usb_set_interface (usb_dev, 0, 0) < 0) 
    {
      printk(KERN_ERR "%s: usb_set_interface() failed!\n", DEVICE_NAME);
      goto reject;
    }
    
    // allocate memory for my device
    if ((dev = (struct pcandev *)kmalloc(sizeof(struct pcandev), GFP_KERNEL)) == NULL)
    {
      printk(KERN_ERR "%s: pcan_usb_plugin - memory allocation failed!\n", DEVICE_NAME);
      goto reject;
    }

    dev->wInitStep = 0;

    u   = &dev->port.usb;

    // init structure elements to defaults
    pcan_soft_init(dev, "usb", HW_USB);

    // preset finish flags
    atomic_set(&u->param_xmit_finished, 0);

    // preset active URB counter
    atomic_set(&u->active_urbs, 0);
  
    // override standard device access functions
    dev->device_open      = pcan_usb_device_open;
    dev->device_release   = pcan_usb_device_release;
    dev->device_write     = pcan_usb_device_write;

    // init process wait queues
    init_waitqueue_head(&dev->read_queue);
    init_waitqueue_head(&dev->write_queue);

    // set this before any instructions, fill struct pcandev, part 1  
    dev->wInitStep   = 0;           
    dev->readreg     = NULL;
    dev->writereg    = NULL;
    dev->cleanup     = pcan_usb_cleanup;
    dev->req_irq     = pcan_usb_req_irq;
    dev->free_irq    = pcan_usb_free_irq;
    dev->open        = pcan_usb_open;
    dev->release     = pcan_usb_release;

    if ((err = assignMinorNumber(dev)))
      goto reject;

    // store pointer to kernel supplied usb_dev
    u->usb_dev          = usb_dev;
    u->ucHardcodedDevNr = (u8)(usb_dev->descriptor.bcdDevice & 0xff);
    u->ucRevision       = (u8)(usb_dev->descriptor.bcdDevice >> 8);
    u->pvXptr           = NULL;

    printk(KERN_INFO "%s: Hardware Revision = %d\n", DEVICE_NAME, u->ucRevision);

    // get endpoint addresses (numbers) and associated max data length
    current_interface_setting = &usb_dev->actconfig->interface->altsetting[usb_dev->actconfig->interface->act_altsetting];
    for (i = 0; i < 4; i++)
    {
      u->Endpoint[i].ucNumber = current_interface_setting->endpoint[i].bEndpointAddress;
      u->Endpoint[i].wDataSz  = current_interface_setting->endpoint[i].wMaxPacketSize;
    }

    spin_lock_init(&u->lock);

    init_waitqueue_head(&u->usb_wait_queue);

    dev->wInitStep = 1;

    // add into list of devices
    list_add_tail(&dev->list, &pcan_drv.devices);  // add this device to the list	      
    dev->wInitStep = 2;
    
    // assign the device as plugged in
    dev->ucPhysicallyInstalled = 1;

    pcan_drv.wDeviceCount++;
    usb_devices++; 
    dev->wInitStep = 3;

    if ((err = pcan_usb_allocate_resources(dev)))
      goto reject;

    printk(KERN_INFO "%s: pcan_usb_plugin() is OK (minor = %d).\n", DEVICE_NAME, dev->nMinor);
    return (void *)dev;

    reject:
    pcan_usb_cleanup(dev);
    
    printk(KERN_ERR "%s: pcan_usb_plugin() failed! (%d)\n", DEVICE_NAME, err);
  }

  return NULL;
}

//****************************************************************************
// is called at plug out of device
//
static void pcan_usb_plugout(struct usb_device *usb_dev, void *drv_context)
{
  struct pcandev *dev = (struct pcandev *)drv_context;

  if (dev)
  {
    DPRINTK(KERN_DEBUG "%s: pcan_usb_plugout(%d)\n", DEVICE_NAME, dev->nMinor);

    // mark this device as plugged out
    dev->ucPhysicallyInstalled = 0;

    // do not remove resources if the device is still in use
    if (!dev->nOpenPaths)
      pcan_usb_cleanup(dev);
  }
}

//****************************************************************************
// propagate getting of serial number to my 'small' interface
//
int  pcan_usb_getSerialNumber(struct pcandev *dev)
{
  if (dev && dev->ucPhysicallyInstalled)
    return pcan_hw_getSNR(dev, &dev->port.usb.dwSerialNumber);
  else
    return -ENODEV;
}

//****************************************************************************
// small interface to rest of driver, only init and deinit
//
int pcan_usb_init(void)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_init()\n", DEVICE_NAME);

	memset (&pcan_drv.usbdrv, 0, sizeof(pcan_drv.usbdrv));

  pcan_drv.usbdrv.probe       = pcan_usb_plugin;
  pcan_drv.usbdrv.disconnect  = pcan_usb_plugout;
  pcan_drv.usbdrv.name        = DEVICE_NAME;
  pcan_drv.usbdrv.minor       = PCAN_USB_MINOR_BASE;
  #ifndef LINUX_22
  pcan_drv.usbdrv.id_table    = pcan_usb_ids;
  #endif

  return usb_register(&pcan_drv.usbdrv);
}

void pcan_usb_deinit(void)
{
  DPRINTK(KERN_DEBUG "%s: pcan_usb_deinit()\n", DEVICE_NAME);

  // unregister usb parts, makes a plugout of registered devices
  usb_deregister(&pcan_drv.usbdrv);
}

 


